from qtop_py.ui.viewport import Viewport
import pytest


def test_defaults():
    viewport = Viewport()
    assert 0 == viewport.h_start
    assert 0 == viewport.h_stop
    assert 0 == viewport.v_start
    assert 0 == viewport.v_stop


def test_after_set_term_size():
    viewport = Viewport()
    viewport.set_term_size(53, 176)
    assert 53 == viewport.v_term_size
    assert 176 == viewport.h_term_size


def test_after_scroll_left():
    viewport = Viewport()
    viewport.set_term_size(53, 176)
    viewport.max_width = 200
    viewport.max_height = 200
    viewport.scroll_left()
    assert 0 == viewport.h_start
    assert 176 == viewport.h_stop  # BUG?? or assert(10 == viewport.h_stop) is a BUG
    assert 0 == viewport.v_start
    assert 53 == viewport.v_stop


def test_after_scroll_right():
    viewport = Viewport()
    viewport.set_term_size(53, 176)
    viewport.max_width = 400
    viewport.max_height = 200
    viewport.scroll_right()
    assert 176 / 2 == viewport.h_start  # corrected behaviour: last element should touch right screen edge, if possible!
    assert 176 / 2 + 176 == viewport.h_stop  # (not scroll endelessly to the right)
    assert 0 == viewport.v_start
    assert 53 == viewport.v_stop


def test_after_scroll_down():
    viewport = Viewport()
    viewport.set_term_size(50, 120)
    viewport.max_width = 400
    viewport.max_height = 200
    viewport.scroll_down()
    assert (0, 120) == (viewport.h_start, viewport.h_stop)
    assert (50, 100) == (viewport.v_start, viewport.v_stop)


def test_after_double_scroll_down():
    viewport = Viewport()
    viewport.set_term_size(50, 120)
    viewport.max_width = 400
    viewport.max_height = 200
    viewport.scroll_down()
    assert (0, 120) == (viewport.h_start, viewport.h_stop)
    assert (50, 100) == (viewport.v_start, viewport.v_stop)
    viewport.scroll_down()
    assert (0, 120) == (viewport.h_start, viewport.h_stop)
    assert (100, 150) == (viewport.v_start, viewport.v_stop)
    viewport.scroll_down()
    assert (0, 120) == (viewport.h_start, viewport.h_stop)
    assert (150, 200) == (viewport.v_start, viewport.v_stop)


def test_after_scroll_right_with_nowhere_to_go():
    viewport = Viewport(hstart=400 - 176)
    viewport.set_term_size(53, 176)
    viewport.max_width = 400
    viewport.max_height = 200
    viewport.scroll_right()
    assert 400 - 176 == viewport.h_start  # OK, shouldn't move!
    assert 400 == viewport.h_stop  # (not scroll endelessly to the right)
    assert 0 == viewport.v_start
    assert 53 == viewport.v_stop


def test_after_scroll_right_when_not_all_space_available_to_the_right():
    viewport = Viewport()
    viewport.set_term_size(53, 176)
    viewport.max_width = 200
    viewport.max_height = 200
    viewport.scroll_right()
    assert 200 - 176 == viewport.h_start  # corrected behaviour: last element should touch right screen edge, if possible!
    assert 200 == viewport.h_stop  # (not scroll endelessly to the right)
    assert 0 == viewport.v_start
    assert 53 == viewport.v_stop


def test_after_scroll_up_no_max_height():
    viewport = Viewport(vstart=400)
    viewport.set_term_size(53, 176)
    viewport.scroll_up()
    assert 0 == viewport.max_height
    assert 0 == viewport.h_start
    assert 176 == viewport.h_stop
    assert 0 == viewport.v_start  # Looks good - didn't change
    assert 53 == viewport.v_stop  # BUG? Should this change now? Why was it 50 before? Did anything really change?


def test_after_scroll_down_no_max_height():
    viewport = Viewport()
    viewport.set_term_size(53, 176)
    viewport.scroll_down()
    assert 0 == viewport.max_height
    assert 0 == viewport.h_start
    assert 176 == viewport.h_stop
    assert 0 == viewport.v_start  # This shouldn't change if there's no maxheight (=empty file)
    assert 53 == viewport.v_stop  # taking up the space of the designated screen


def test_after_scroll_up():
    viewport = Viewport(vstart=276)
    viewport.set_term_size(53, 176)
    viewport.max_height = 300
    viewport.scroll_up()
    assert 300 == viewport.max_height
    assert 0 == viewport.h_start
    assert 176 == viewport.h_stop
    assert 276 - 53 == viewport.v_start
    assert 276 == viewport.v_stop


def test_after_scroll_down_scroll_up():
    viewport = Viewport()
    viewport.set_term_size(53, 176)
    viewport.max_height = 300
    viewport.scroll_down()
    assert 300 == viewport.max_height
    assert 0 == viewport.h_start
    assert 176 == viewport.h_stop
    assert 53 == viewport.v_start  # This scrolling logic looks ok
    assert 106 == viewport.v_stop

    viewport.scroll_up()  # Now we can scroll up again

    assert 300 == viewport.max_height
    assert 0 == viewport.h_start
    assert 176 == viewport.h_stop
    assert 0 == viewport.v_start
    assert 53 == viewport.v_stop


def test_after_reset_to_starting_position():
    viewport = Viewport(400, 500)
    viewport.set_term_size(53, 176)
    viewport.max_width = 500
    viewport.max_height = 500
    viewport.reset_display()
    assert 0 == viewport.v_start
    assert 0 == viewport.h_start
    assert 53 == viewport.v_stop
    assert 176 == viewport.h_stop


def test_scroll_far_right_attaches_to_right_screen_edge():
    viewport = Viewport(200, 200)
    viewport.set_term_size(53, 176)
    viewport.max_width = 400
    viewport.max_height = 400
    viewport.scroll_far_right()
    assert 400 - 176 == viewport.h_start
    assert 400 == viewport.h_stop
    assert 200 == viewport.v_start
    assert 200 + 53 == viewport.v_stop


"""
This is a quick'n'clean way to run many edge cases without re-writing the whole bloody initialisation every time!!
"""
@pytest.mark.parametrize('init_vstart, init_hstart, term_size, max_matrix_dim, expected',
    (
        (0, 0, [53, 176], (200, 200), (0, 53, 24, 200)),  # test1
        (0, 100, [53, 176], (400, 400), (0, 53, 188, 364)),  # test2 etc
        (0, 100, [53, 176], (400, 200), (0, 53, 100, 276)),
        (0, 199, [53, 176], (400, 200), (0, 53, 199, 375)),
    ),
)
def test_after_scroll_right(init_vstart, init_hstart, term_size, max_matrix_dim, expected):
    viewport = Viewport(init_vstart, init_hstart)
    viewport.set_term_size(*term_size)
    viewport.max_width = max_matrix_dim[1]
    viewport.max_height = max_matrix_dim[0]
    viewport.scroll_right()
    # corrected behaviour: last element should touch right screen edge, if possible!
    assert expected == (viewport.v_start, viewport.v_stop, viewport.h_start, viewport.h_stop)


@pytest.mark.parametrize('init_vstart, init_hstart, term_size, max_matrix_dim, expected',
     (
         (0, 0, [30, 120], (200, 200), (170, 200, 0, 120)),  # from top to bottom
         (40, 50, [30, 120], (200, 200), (170, 200, 50, 170)),  # from random to bottom
         (170, 50, [30, 120], (200, 200), (170, 200, 50, 170)),  # from bottom to bottom
     ),
)
def test_after_scroll_bottom(init_vstart, init_hstart, term_size, max_matrix_dim, expected):
    viewport = Viewport(init_vstart, init_hstart)
    viewport.set_term_size(*term_size)
    viewport.max_width = max_matrix_dim[1]
    viewport.max_height = max_matrix_dim[0]
    viewport.scroll_bottom()
    assert expected == (viewport.v_start, viewport.v_stop, viewport.h_start, viewport.h_stop)



import pytest
from qtop_py.yaml_parser import (fix_config_list, get_line, convert_dash_key_in_dict, parse, read_yaml_config_block,
                         process_line, process_code, safe_load, load_all, get_yaml_key_part)


@pytest.mark.parametrize('fin, t',
    (
        (
            ['scheduler: sge'],
            [0, "scheduler:", "sge"]
        ),
        (
            ['schedulers'],
            [0, "schedulers"]
        ),
        (
            ['  pbs'],
            [1, "pbs"]
        ),
        (
            ["   - r'moonshot'"],
            [2, "-", "r'moonshot'"]
        ),
        (
            [" - '\w*cms048': Blue"],
            [1, "-", "'\\w*cms048': Blue"]
        ),
        (
            ["term_size: [53, 176]"],
            [0, "term_size:", "[53, 176]"]
        ),
    ),
)
def test_detect_indentation_level(fin, t):
    get_lines = get_line(fin, verbatim=False)
    actual_t = next(get_lines)
    assert actual_t == t

# @pytest.mark.skipif(True, reason="No special reason")
@pytest.mark.parametrize('fin, t',
    (
        (
            [
                '  oar:\n',
                '    oarnodes_s_file: oarnodes_s_Y.txt\n'
            ],
            [
                [1, "oar:"],
                [1, 'oarnodes_s_file:', 'oarnodes_s_Y.txt']
            ]
        ),
        (
            [
                '    oarstat_file: oarstat.txt\n',
                '  sge'
            ],
            [
                [2, "oarstat_file:", "oarstat.txt"],
                [-1, "sge"]
            ]
        ),
        (
            [
                '    sge_file: qstat.F.xml.stdout',
                '\n',
                'faster_xml_parsing: False'
            ],
            [
                [2, 'sge_file:', 'qstat.F.xml.stdout'],
                [-2],
                [0, 'faster_xml_parsing:', 'False'],
            ]
        )
    ),
)
def test_get_more_lines(fin, t):
    get_lines = get_line(fin)
    for idx, line in enumerate(fin):
        actual_t = next(get_lines)
        assert actual_t == t[idx]

# @pytest.mark.parametrize('fin, line, get_lines, key_container, parent_container',  # get_line(fin)
#      (
#          (
#              ['testkey: testvalue'], [0, 'testkey', 'testvalue'], get_line(['testkey: testvalue']), {'testkey': 'testvalue'}, 'testvalue'
#          ),
#          (
#              ['testkey'], [0, 'testkey'], get_line(['testkey']), {'testkey': {}}, {}
#          ),
#          (
#              ['testkey: [testvalue]'], [0, 'testkey', '[testvalue]'], get_line(['testkey: [testvalue]']), {'testkey': ['testvalue']}, ['testvalue']
#          ),
#          # (
#          #     ['testkey: |'], [0, 'testkey', '|'], get_line(['testkey: |']), {'testkey': {}}, {}
#          # ),
#          (
#              ['- testkey'], [0, '-', 'testkey'], get_line(['- testkey']), {'-': ['testkey']}, 'testkey'
#          ),
#          (
#              ['- testkey: testvalue'], [0, '-', 'testkey: testvalue'], get_line(['- testkey: testvalue']), {'-': [{'testkey': 'testvalue'}]}, 'testvalue'
#          ),
#          (
#              ['- testkey: [testvalue]'], [0, '-', 'testkey: [testvalue]'], get_line(['- testkey: [testvalue]']), {'-': [{'testkey': ['testvalue']}]}, ['testvalue']
#          ),
#          (
#              ['- testvalue'], [0, '-', 'testvalue'], get_line(['- testvalue']), {'-': ['testvalue']}, 'testvalue'
#          ),
#      )
# )
# def test_process_qstat_line(line, fin, get_lines, key_container, parent_container):  # parent_container, container_stack, stack):
#     assert process_line(line, fin, get_lines, parent_container) == (key_container, parent_container)


# @pytest.mark.current
@pytest.mark.parametrize('line_in, fin, block_out, line_out',
     (
         (
             [0],
"""testkey1: testvalue1
testkey2: testvalue2
testkey3: testvalue3
""".split('\n'),
             {'testkey1': 'testvalue1', 'testkey2': 'testvalue2', 'testkey3': 'testvalue3'},
             [0],
         ),
         (
             [0],

"""testkey4:
  testkey5:
    testkey6: value6
    testkey7: value7
""".split('\n'),
             {'testkey4': {'testkey5': {'testkey6': 'value6', 'testkey7': 'value7'}}},
             [-2],
         ),
         (
             [0],
"""testkey8:
  - testkey9:
     testkey10: value10
  - testkey11:
     testkey12: value12
  - testkey13: value13
  - value14
""".split('\n'),
             {'testkey8':
                  {
                    '-':
                       [
                        {'testkey9': {'testkey10': 'value10'}},
                        {'testkey11': {'testkey12': 'value12'}},
                        {'testkey13': 'value13'},
                        'value14'
                       ]
                  }
             },
             [-1],
         ),
         (
             [0],
"""testkey36:
  - testkey37: testvalue37
    testkey38: testvalue38
  - testkey39: testvalue39
    testkey40: testvalue40
  - testkey41: testvalue41
    testkey42: testvalue42
""".split('\n'),
             {'testkey36':
                  {
                    '-':
                       [
                        {'testkey37': 'testvalue37', 'testkey38': 'testvalue38'},
                        {'testkey39': 'testvalue39', 'testkey40': 'testvalue40'},
                        {'testkey41': 'testvalue41', 'testkey42': 'testvalue42'},
                       ]
                  }
             },
             [-2],
         ),
         (
             [0],
"""testkey15:
    - testkey16:
        testkey17: testvalue17
        testkey18: [testvalue18]
""".split('\n'),
             {'testkey15': {'-': [{'testkey16': {'testkey17': 'testvalue17', 'testkey18': ['testvalue18']}}]}},
             [-8],
         ),
         (
             [0],
"""testkey19:
 - testkey20:
     testkey21: testvalue21
""".split('\n'),
             {'testkey19': {'-': [{'testkey20': {'testkey21': 'testvalue21'}}]}},
             [-5],
         ),
         (
             [0],
"""testkey22:  # order should be from more generic-->more specific
 - testkey23: testvalue23
 - testkey24: testvalue24
 - testkey25: testvalue25
""".split('\n'),
             {'testkey22': {'-': [{'testkey23': 'testvalue23'}, {'testkey24': 'testvalue24'}, {'testkey25': 'testvalue25'}]}},
             [-1],
         ),
         (
             [0],
"""testkey26:
 - testvalue27
 - testvalue28
 - testvalue29
""".split('\n'),
             {
                  'testkey26':
                      {'-':
                           ['testvalue27', 'testvalue28', 'testvalue29']
                      }
             },
             [-1],
         ),
         (
             [0],
"""testkey31:
 - testkey32: [testvalue32, testvalue32b]
 - testkey33:
    - testvalue34
    - testvalue35
""".split('\n'),
             {'testkey31':
                  {'-':
                       [
                           {'testkey32': ['testvalue32', 'testvalue32b']},
                           {'testkey33':
                                {'-':
                                     ['testvalue34', 'testvalue35']
                                 }
                            }
                       ]
                  }
             },
             [-2],
         ),
         (
             [0],
"""state_abbreviations:
  pbs:
    Q: queued_of_user
    E: exiting_of_user
    W: waiting_of_user
  oar:
    E: Error
    F: Finishing
    S: cancelled_of_user
  sge:
    r: running_of_user
    E: exiting_of_user  # not real
    qw: queued_of_user
""".split('\n'),
             {'state_abbreviations': {'pbs': {'Q': 'queued_of_user', 'E': 'exiting_of_user', 'W': 'waiting_of_user'},
                                       'oar': {'E': 'Error', 'F': 'Finishing', 'S': 'cancelled_of_user'}, 'sge': {'r': 'running_of_user', 'E': 'exiting_of_user', 'qw': 'queued_of_user'}}},
             [-2],
         ),
#          (
#              [0],
#
# """user_sort: |
#     lambda d: (
#     d['np'],
#     ord(d['domainname'][0]),
#     len(d['domainname'].split('.', 1)[0].split('-')[0]),
#     )
# """.split('\n'),
#              {'user_sort': "lambda d: (\nd['np'],\nord(d['domainname'][0]),\nlen(d['domainname'].split('.', 1)[0].split('-')[0]),\n)"},
#              [-1],
#          ),
     ), ids=[
        "3dicts",
        "dict2items_inside_doubly_nested_dict",
        "listof_3dicts_and_a_value_inside_dict",
        "list_dicts",
        "1block",
        "2block",
        "3block",
        "lov_in_dict",
        "nested_lists",
        "state_abbreviations",
        # "code"
    ]
)
def test_read_yaml_config_block(line_in, line_out, fin, block_out):
    get_lines = get_line(fin)
    assert read_yaml_config_block(line_in, fin, get_lines) == (block_out, line_out)


@pytest.mark.parametrize('fin, code',
     (
        (
            [
                "lambda d: (",
                "d['np'],",
                "ord(d['domainname'][0]),",
                "len(d['domainname'].split('.', 1)[0].split('-')[0]),",
                "int(d['domainname'].split('.', 1)[0].split('-')[1]),",
            ],
            "lambda d: (\nd['np'],\nord(d['domainname'][0]),\nlen(d['domainname'].split('.', 1)[0].split('-')[0]),\nint(d['domainname'].split('.', 1)[0].split('-')[1]),"
        ),
     )
)
def test_process_code(fin, code):
    assert process_code(fin) == code

# @pytest.mark.current
@pytest.mark.parametrize('dict_a, dict_b',
    (
        (
                {'filt': 'a', 'filtb': 'b'},  # nothing should change here
                {'filt': 'a', 'filtb': 'b'},
        ),
        (
            {'testkey22':
                {'-':
                     [
                         {'testkey23': 'testvalue23'},
                         {'testkey24': 'testvalue24'},
                         {'testkey25': 'testvalue25'}
                     ]
                }
            },
            {'testkey22':
                     [
                         {'testkey23': 'testvalue23'},
                         {'testkey24': 'testvalue24'},
                         {'testkey25': 'testvalue25'}
                     ]
            }
        ),
        # (
        #     {'filt':
        #          {'-':
        #               [
        #                   {'list_name': [1, 2, 3]},
        #                   {'list_pattern': {'-': ['p5', 'moonshot']}}
        #               ]
        #          }
        #     },
        #     {'filt':
        #          [
        #              {'list_name': [1, 2, 3]},
        #              {'list_pattern': ['p5', 'moonshot']}
        #          ]
        #     }
        # ),
    )
)
def test_convert_dash_key_in_dict(dict_a, dict_b):
    # pass
    assert convert_dash_key_in_dict(dict_a) == dict_b



__author__ = 'sfranky'
try:
    import ujson as json
except ImportError:
    import json
import logging
import sys
from qtop_py.serialiser import StatExtractor, GenericBatchSystem
from xml.etree import ElementTree as etree
import qtop_py.fileutils as fileutils


class SGEStatExtractor(StatExtractor):
    def __init__(self, config, options):
        StatExtractor.__init__(self, config, options)

    def extract_qstat(self, orig_file):
        all_values = list()
        try:
            tree = etree.parse(orig_file)
        except etree.ParseError:
            logging.critical("This is an XML parse error (??)")
            raise
        except IOError:
            raise
        except:
            print "File %(filename)s does not appear to contain a proper XML structure. Exiting.." % {"filename": orig_file}
            raise
        else:
            root = tree.getroot()

        for queue_elem in root.findall('queue_info/Queue-List'):
            queue_name_elems = queue_elem.findall('resource')
            for queue_name_elem in queue_name_elems:
                if queue_name_elem.attrib.get('name') == 'qname':
                    queue_name_elem.text = queue_name_elem.text if not self.options.ANONYMIZE else self.anonymize(queue_name_elem.text, 'qs')
                    break
            else:
                raise ValueError("No such queue name")

            try:
                all_values = self._extract_job_info(all_values, queue_elem, 'job_list', queue_name=queue_name_elem.text)
            except ValueError:
                logging.info('No jobs found in XML file!')

        job_info_elem = root.find('./job_info')
        if job_info_elem is None:
            logging.debug('No pending jobs found!')
        else:
            try:
                all_values = self._extract_job_info(all_values, job_info_elem, 'job_list', queue_name='Pending')
            except ValueError:
                logging.info('No jobs found in XML file!')

        if self.options.SAMPLE >= 1:
            tree.write(orig_file)  # TODO anonymize rest of the sensitive information within xml file
        return all_values

    def _extract_job_info(self, all_values, elem, elem_text, queue_name):
        """
        inside elem, iterates over subelems named elem_text and extracts relevant job information
        TODO: check difference between extract_job_info and _extract_job_info
        """
        for subelem in elem.findall(elem_text):
            qstat_values = dict()
            qstat_values['JobId'] = subelem.find('./JB_job_number').text
            qstat_values['UnixAccount'] = subelem.find('./JB_owner').text \
                if not self.options.ANONYMIZE else self.anonymize(subelem.find('./JB_owner').text, 'users')
            qstat_values['S'] = subelem.find('./state').text
            qstat_values['Queue'] = queue_name
            all_values.append(qstat_values)
        if not all_values:
            raise ValueError('No jobs found in XML file!')

        return all_values


class SGEBatchSystem(GenericBatchSystem):

    @staticmethod
    def get_mnemonic():
        return "sge"

    def __init__(self, scheduler_output_filenames, config, options):
        self.sge_file = scheduler_output_filenames.get('sge_file')
        self.config = config
        self.options = options
        self.sge_stat_maker = SGEStatExtractor(self.config, self.options)

    def get_queues_info(self):
        logging.debug("Parsing tree of %s" % self.sge_file)
        fileutils.check_empty_file(self.sge_file)
        anonymize = self.sge_stat_maker.anonymize_func()

        tree = self._get_xml_tree(self.sge_file)
        root = tree.getroot()

        qstatq_list = self._extract_queues('queue_info/Queue-List', root)

        total_running_jobs = sum([d['run'] for d in qstatq_list])
        logging.info('Total running jobs found: %s' % total_running_jobs)

        for d in qstatq_list:
            d['run'] = str(d['run'])
            d['queued'] = str(d['queued'])

        total_queued_jobs = self._get_total_queued_jobs('job_info/job_list', root)

        qstatq_list.append({'run': '0', 'queued': total_queued_jobs, 'queue_name': 'Pending', 'state': 'Q', 'lm': '0'})
        logging.debug('qstatq_list contains %s elements' % len(qstatq_list))
        # TODO: check validity. 'state' shouldnt just be 'Q'!
        logging.debug("Closing %s" % self.sge_file)

        return total_running_jobs, int(eval(str(total_queued_jobs))), qstatq_list

    def get_worker_nodes(self, job_ids, job_queues, options):
        logging.debug('Parsing tree of %s' % self.sge_file)
        anonymize = self.sge_stat_maker.anonymize_func()

        with open(self.sge_file, 'rb') as fin:
            tree = etree.parse(fin)

        root = tree.getroot()
        existing_wns = list()
        existing_node_names = set()

        for queue_elem in root.findall('queue_info/Queue-List'):
            worker_node = self._get_host_qname_np(queue_elem, anonymize)
            worker_node['state'] = self._get_state(queue_elem)

            if worker_node['domainname'] not in existing_node_names:
                job_ids, _, _ = self._extract_job_info(queue_elem, 'job_list')
                worker_node['core_job_map'] = dict((idx, job_id) for idx, job_id in enumerate(job_ids))
                worker_node['existing_busy_cores'] = len(worker_node['core_job_map'])
                worker_node['np'] = max(int(worker_node['np']), len(worker_node['core_job_map']))

                existing_node_names.update([worker_node['domainname']])
                existing_wns.append(worker_node)
            else:
                for existing_wn in existing_wns:
                    if worker_node['domainname'] != existing_wn['domainname']:
                        continue

                    job_ids, _, _ = self._extract_job_info(queue_elem, 'job_list')
                    core_jobs = dict((idx, job_id) for idx, job_id in enumerate(job_ids, existing_wn['existing_busy_cores']))
                    existing_wn['core_job_map'].update(core_jobs)
                    existing_wn['existing_busy_cores'] = len(existing_wn['core_job_map'])
                    # don't change the node state to free.
                    # Just keep the state reported in the last queue mentioning the node.
                    existing_wn['state'] = (worker_node['state'] == '-') and existing_wn['state'] or worker_node['state']
                    existing_wn['qname'].update(worker_node['qname'])
                    existing_wn['np'] = max(int(existing_wn['np']), len(existing_wn['core_job_map']))
                    break

        logging.debug('Closing %s' % self.sge_file)
        logging.info('existing_wns contains %s entries' % len(existing_wns))
        for existing_wn in existing_wns:
            existing_wn['qname'] = list(existing_wn['qname'])

        return existing_wns

    def get_jobs_info(self):
        job_ids, usernames, job_states, queue_names = [], [], [], []

        all_values = self.sge_stat_maker.extract_qstat(self.sge_file)
        # TODO: needs better glueing
        for qstat in all_values:
            job_ids.append(str(qstat['JobId']))
            usernames.append(qstat['UnixAccount'])
            job_states.append(qstat['S'])
            queue_names.append(qstat['Queue'])

        logging.debug('job_ids, usernames, job_states, queue_names lengths: '
                      '%(job_ids)s, %(usernames)s, %(job_states)s, %(queue_names)s'
                      % {
                          "job_ids": len(job_ids),
                          "usernames": len(usernames),
                          "job_states": len(job_states),
                          "queue_names": len(queue_names)
                      }
                      )
        return job_ids, usernames, job_states, queue_names

    def _extract_job_info(self, elem, elem_text):
        """
        inside elem, iterates over subelems named elem_text and extracts relevant job information
        TODO: check difference between extract_job_info and _extract_job_info
        """
        job_ids, usernames, job_states = [], [], []
        for subelem in elem.findall(elem_text):
            job_ids.append(subelem.find('./JB_job_number').text)
            usernames.append(subelem.find('./JB_owner').text)
            job_states.append(subelem.find('./state').text)
        return job_ids, usernames, job_states

    def _get_xml_tree(self, xml_file):
        with open(xml_file, mode='rb') as fin:
            try:
                tree = etree.parse(fin)
            except etree.ParseError:
                logging.critical("Something happened during the parsing of the XML file. Exiting...")
            except:
                logging.debug("XML file state %s" % fin)
                logging.debug("thinking...")
                sys.exit(1)
        return tree

    def _get_host_qname_np(self, queue_elem, anonymize):
        worker_node = dict()
        count = 0
        try:
            slots_used = int(queue_elem.find('./slots_used').text)
        except AttributeError:
            slots_used = 0

        resources = queue_elem.findall('resource')
        for resource in resources:
            if resource.attrib.get('name') == 'hostname':
                worker_node['domainname'] = resource.text if not self.options.ANONYMIZE else anonymize(resource.text, 'wns')
                count += 1
            elif resource.attrib.get('name') == 'qname':
                qname = resource.text
                if not slots_used:
                    worker_node['qname'] = set()
                else:
                    # if slots are reportedly used, the queue will be displayed even if no actual running jobs exist
                    worker_node['qname'] = set([qname]) if not self.options.ANONYMIZE else set([anonymize(qname, 'qs')])
                count += 1
            elif resource.attrib.get('name') == 'num_proc':
                worker_node['np'] = resource.text
                count += 1
            if count == 3:
                break
        else:
            # out of the 3, np information is the most likely to be missing from a node
            worker_node['np'] = 0

        return worker_node

    def _get_state(self, queue_elem):
        try:
            _state = queue_elem.find('state').text
        except AttributeError:
            _state = '-'
        finally:
            return _state

    def _extract_queues(self, xpath, root):
        qstatq_list = []
        for queue_elem in root.findall(xpath):

            queue_names = queue_elem.findall('resource')
            for _queue_name in queue_names:
                if _queue_name.attrib.get('name') == 'qname':
                    queue_name = _queue_name.text if not self.options.ANONYMIZE else anonymize(_queue_name.text, 'qs')
                    break
            else:
                raise ValueError("No such resource")

            for exist_d in qstatq_list:
                if queue_name == exist_d['queue_name']:

                    jobs = queue_elem.findall('job_list')
                    run_count = 0
                    for _run in jobs:
                        if _run.attrib.get('state') == 'running':
                            run_count += 1
                    exist_d['run'] += run_count
                    break
            else:  # first instance of queue in the xml
                d = dict()
                d['queue_name'] = queue_name
                try:
                    d['state'] = queue_elem.find('./state').text
                except AttributeError:
                    d['state'] = '?'
                except:
                    raise

                job_lists = queue_elem.findall('job_list')
                run_count = 0
                for _run in job_lists:
                    if _run.attrib.get('state') == 'running':
                        run_count += 1
                d['run'] = run_count
                d['lm'] = 0
                d['queued'] = 0
                qstatq_list.append(d)

        return qstatq_list

    def _get_total_queued_jobs(self, xpath, root):
        total_queued_jobs_elems = root.findall(xpath)
        pending_count = 0
        for job in total_queued_jobs_elems:
            if job.attrib.get('state') == 'pending':
                pending_count += 1
        total_queued_jobs = str(pending_count)
        logging.info('Total queued jobs found: %s' % total_queued_jobs)

        return total_queued_jobs



import os
import logging


def fix_config_list(config_list):
    """
    transforms a list of the form ['a, b'] to ['a', 'b']
    """
    if not config_list:
        return []
    t = config_list
    item = t[0]
    list_items = item.split(',')
    return [nr.strip() for nr in list_items]


def get_line(fin, verbatim=False, SEPARATOR=None, DEF_INDENT=2):
    """
    Yields a list per line read, of the following form:
    [indentation_change, line up to first space, line after first space if exists]
    Comment lines are omitted.
    Lines where comments exist at the end are stripped off their comment.
    Indentation is calculated with respect to the previous line.
    1: line is further indented
    0: same indentation
    -1: line is unindent
    Empty lines only return the indentation change.
    Where the line splits depends on SEPARATOR (default is first space)

    DEF_INDENT is how many spaces is an indent by default.
    e.g. qtop config file uses 2 spaces, oarnodes_s_y uses 4
    """
    indent = 0
    indenter = {
        0: 0,
        DEF_INDENT / 2: 1,
        DEF_INDENT: 1,
        3 * int(float(DEF_INDENT) / 2): 2,
        2 * DEF_INDENT: 2,
        -DEF_INDENT / 2: -1,
        -DEF_INDENT: -1,
        -3 * int(float(DEF_INDENT) / 2): -2,
        - 2 * DEF_INDENT: -2,
    }
    for line in fin:
        if line.lstrip().startswith('#') or line.strip() == '---':
            continue
        elif ' #' in line and not line.endswith('#\n'):
            line = line.split(' #', 1)[0]

        prev_indent = indent
        indent = len(line) - len(line.lstrip(' '))
        diff = indent - prev_indent
        try:
            d_indent = indenter[diff]
        except KeyError:
            d_indent = diff

        line = line.rstrip()
        list_line = verbatim and [d_indent, line] or [d_indent] + line.split(None or SEPARATOR, 1)

        if len(list_line) > 1:
            if list_line[1].startswith(('"', "'")):
                list_line[1] = list_line[1][1:-1]
        else:
            pass
        yield list_line


def convert_dash_key_in_dict(d):
    """
    takes a dict of the form {'-': [...]} and converts it to [...]
    """
    try:
        assert isinstance(d, dict)
    except AssertionError:
        return d  # TODO: Maybe this should fail, not be muted

    for key_out in d:
        if not (isinstance(d[key_out], dict) or len(d[key_out]) == 1):
            continue
        try:
            for key_in in d[key_out]:
                if key_in == '-' and key_out != 'state':
                    d[key_out] = d[key_out][key_in]
                # elif key_in == '-' and key_out == 'state':
                #     d[key_out] = eval(d[key_out])
                #     break
        except TypeError:
            return d
        except IndexError:
            continue
    return d


def parse(fn, DEF_INDENT=2):
    raw_key_values = {}
    with open(fn, mode='r') as fin:
        try:
            assert os.stat(fn).st_size != 0
        except AssertionError:
            logging.critical('File %s is empty!! Exiting...\n' % fn)
            raise
        except IOError:
            raise
        logging.debug('File state before parse: %s' % fin)
        get_lines = get_line(fin, DEF_INDENT=DEF_INDENT)  # TODO: weird
        line = next(get_lines)
        while line:
            block, line = read_yaml_config_block(line, fin, get_lines)
            block = convert_dash_key_in_dict(block)
            for k in block:
                block[k] = convert_dash_key_in_dict(block[k])
            raw_key_values.update(block)

    logging.debug('File state after parse: %s' % fin)
    a_dict = dict([(key, value) for key, value in raw_key_values.items()])
    return a_dict


def read_yaml_config_block(line, fin, get_lines):
    block = dict()
    parent_container = block
    open_containers = list()
    open_containers.append(block)

    # if len(line) > 1:  # non-empty line
    #     key_value, parent_container = process_line(line, fin, get_lines, parent_container)
    #     for (k, v) in key_value.items():
    #         block[k] = v

    while len(line) == 1:  # skip empty lines
        try:
            line = next(get_lines)
        except StopIteration:  # EO(config)F
            return {}, ''

    while len(line) > 1:  # as long as a blank line is not reached (i.e. block is not complete)
        # if line[0] == 0 or (line[0] != 0 and line[1] == '-'):  # same level
        # key_value used below belongs to previous line. It will work for first block line because of short circuit logic
        if line[0] == 0 \
                or (line[0] == 1 and (key_value.keys()[0] == '-'))\
                or (line[0] == -1 and line[1] == '-'):  # same level or entry level
            key_value, container = process_line(line, fin, get_lines, parent_container)
            for k in key_value:
                pass  # assign dict's sole key to k
            if parent_container == {} or '-' not in parent_container:
                parent_container[k] = key_value[k]
            elif '-' in parent_container and '-' not in key_value:
                last_item = parent_container['-'].pop()
                key_value.update(last_item)
                parent_container['-'].append(key_value)
            else:
                parent_container.setdefault(k, []).extend(key_value[k])  # was waiting for a list, but a str came in!
            if container == {}:
                open_containers.append(container)
                parent_container = open_containers[-1]  # point towards latest container (key_value's value)

        elif (line[0] == 1) or (line[0] > 1):  # go down one level
            key_value, container = process_line(line, fin, get_lines, parent_container)
            for k in key_value:
                pass
            # if container == {}:  # up parent container with new value
            if parent_container == {}:  # above it is a key waiting to be filled with values
                parent_container[k] = key_value[k]
            else:
                parent_container.setdefault(k, []).append(key_value[k]) if isinstance(key_value[k], str) else \
                    parent_container.setdefault(k, []).extend(key_value[k])
            if container == {}:
                open_containers.append(container)
                parent_container = open_containers[-1]  # point towards latest container (key_value's value)

        elif line[0] == -2 and line[1] == '-':  # go up two levels
            key_value, container = process_line(line, fin, get_lines, parent_container)
            len(open_containers) > 1 and open_containers.pop() or None
            for k in key_value:
                pass  # assign dict's sole key to k
            if open_containers[-1].get('-'):
                open_containers[-1].setdefault('-', []).extend(key_value[k])
            else:
                open_containers[-1][k] = key_value[k]
            if container == {}:
                open_containers.append(container)
                parent_container = open_containers[-1]  # point towards latest container (key_value's value)
            else:
                parent_container = open_containers[-1]

        elif line[0] == -1:  # go up one level
            key_value, container = process_line(line, fin, get_lines, parent_container)
            len(open_containers) > 1 and open_containers.pop() or None
            for k in key_value:
                pass  # assign dict's sole key to k
            if open_containers[-1].get('-'):
                open_containers[-1].setdefault('-', []).extend(key_value[k])
            else:
                open_containers[-1][k] = key_value[k]
            if container == {}:
                open_containers.append(container)
                parent_container = open_containers[-1]  # point towards latest container (key_value's value)
            else:
                parent_container = open_containers[-1]

        try:
            line = next(get_lines)
        except StopIteration:
            return block, ''
        else:
            if line[-1] == '...':
                return block, line
    return block, line


def process_line(list_line, fin, get_lines, parent_container):
    key = list_line[1]

    if len(list_line) == 2:  # key-only, so what's in the line following should be written in a new container
        container = {}
        return {key.rstrip(':'): container}, container

    elif len(list_line) == 3:
        container = list_line[2]

        if container.endswith(':'):  # key: '-'           - testkey:
            parent_key = key
            key = container
            new_container = {}
            return {parent_key: [{key.rstrip(':'): new_container}]}, new_container  #list

        elif ': ' in container:  # key: '-'               - testkey: testvalue
            parent_key = key
            key, container = container.split(None, 1)
            # container = [container[1:-1]] if container.startswith('[') else container
            container = container[1:-1].split(', ') if container.startswith('[') else container
            container = "" if container in ("''", '""') else container
            if len(container) == 1 and isinstance(container, list) and isinstance(container[0], str):
                try:
                    container = list(eval(container[0]))
                except NameError:
                    pass
            return {'-': [{key.rstrip(':'): container}]}, container  #list

        elif container.endswith('|'):
            container = process_code(fin)
            return {key.rstrip(':'): container}, parent_container

        else:  # simple value
            if key == '-':  # i.e.  - testvalue
                return {'-': [container]}, container  # was parent_container******was :[container]}, container
            else:  # i.e. testkey: testvalue
                container = [container[1:-1]] if container.startswith('[') else container  #list
                if len(container) == 1 and isinstance(container, list) and isinstance(container[0], str):
                    try:
                        container = list(eval(container[0]))
                    except NameError:
                        pass
                elif container.startswith("'") and container.endswith("'"):
                    container = eval(container)
                return {key.rstrip(':'): container}, container  # was parent_container#str
    else:
        raise ValueError("Didn't anticipate that!")


def process_code(fin):
    get_code = get_line(fin, verbatim=True)
    # line = next(get_code)
    line = next(get_code)
    code = []
    while line[0] > -1:
        code.append(' ' + line[-1])
        try:
            line = next(get_code)
        except StopIteration:
            break
    return '\n'.join([c.strip() for c in code]).strip()


def safe_load(fin, DEF_INDENT=2):
    a_dict = parse(fin, DEF_INDENT)
    logging.debug("YAML dict length: %s" % len(a_dict))
    return a_dict


def load_all(fin):
    list_of_dicts = []
    get_lines = get_line(fin)
    while True:
        try:
            line = next(get_lines)
        except StopIteration:
            break
        block, line = read_yaml_config_block(line, fin, get_lines)
        block = convert_dash_key_in_dict(block)
        list_of_dicts.append(block)

    return list_of_dicts


def get_yaml_key_part(config, scheduler, outermost_key):
    """
    only return the list items of the yaml outermost_key if a yaml key subkey exists
    (this signals a user-inserted value)
    """
    # e.g. outermost_key = 'workernodes_matrix'
    for part in config[outermost_key]:
        part_name = [i for i in part][0]
        part_options = part[part_name]
        yaml_key = part_options.get('yaml_key')
        # if no systems line exists, all systems are supported, and thus the current
        systems = fix_config_list(part_options.get('systems', [scheduler]))
        if yaml_key:
            yield yaml_key, part_name, systems



from SimpleHTTPServer import SimpleHTTPRequestHandler
import SocketServer
from multiprocessing import Process, Queue
from threading import Thread
import os


class Web(object):
    def __init__(self, initial_cwd):
        self.web_dir = initial_cwd + "/web"
        self.started = False
        self.filename = None

    def start(self):
        self.started = True
        self.q = Queue()

        # This will run in another process which will communicate through
        # the `q` Queue.
        def http_server(q):
            os.chdir(self.web_dir)

            class SharedVars(object):
                def __init__(self):
                    self.filename = ""

            shared = SharedVars()

            class MyHandler(SimpleHTTPRequestHandler):
                ''' Overriding the default behaviour just for api.json'''

                def log_message(self, format, *args):
                    '''quiet!'''
                    pass

                def do_GET(self):
                    # Handle the case of api.json...
                    if self.path == "/api.json" and shared.filename:
                        with open(shared.filename, 'r') as f:
                            read_data = f.read()

                        self.send_response(200)
                        self.send_header("Content-type",
                                         "Content-Type: application/json")
                        self.send_header("Content-length", len(read_data))
                        self.end_headers()
                        self.wfile.write(read_data)
                    else:
                        # Default behaviour
                        SimpleHTTPRequestHandler.do_GET(self)

            class MyTCPServer(SocketServer.TCPServer):
                # Undocumented. Getting rid of "address already in use" errors
                allow_reuse_address = True

            # Hardcoded... who cares right now.
            PORT = 8080
            httpd = MyTCPServer(("", PORT), MyHandler)
            thread = Thread(target=httpd.serve_forever, args=())
            thread.start()

            # This is the main loop of this process. I supports two commands...
            while True:
                command = q.get()

                # The "stop" command terminates the server and exits the loop
                if command == "stop":
                    httpd.shutdown()
                    httpd.server_close()
                    thread.join()
                    break

                # The "setfilename" command sets the file where "api.json"
                # points to.
                elif command.startswith("setfilename "):
                    # Setting the shared filename that will
                    # be picked up by server's thread
                    shared.filename = command.split()[1]

        self.server = Process(target=http_server, args=(self.q, ))
        self.server.start()

    def set_filename(self, filename):
        if self.started:
            self.q.put("setfilename " + filename)

    def stop(self):
        if self.started:
            self.q.put("stop")
            self.server.join()



"""
Taken from: http://code.activestate.com/recipes/576611-counter-class/
This module was added to ensure compatibility with older versions of python
which do not include the named tuple structure from the collections module.
"""


from operator import itemgetter
from heapq import nlargest
from itertools import repeat, ifilter


class Counter(dict):
    '''Dict subclass for counting hashable objects.  Sometimes called a bag
    or multiset.  Elements are stored as dictionary keys and their counts
    are stored as dictionary values.

    >>> Counter('zyzygy')
    Counter({'y': 3, 'z': 2, 'g': 1})

    '''

    def __init__(self, iterable=None, **kwds):
        '''Create a new, empty Counter object.  And if given, count elements
        from an input iterable.  Or, initialize the count from another mapping
        of elements to their counts.

        >>> c = Counter()                           # a new, empty counter
        >>> c = Counter('gallahad')                 # a new counter from an iterable
        >>> c = Counter({'a': 4, 'b': 2})           # a new counter from a mapping
        >>> c = Counter(a=4, b=2)                   # a new counter from keyword args

        '''
        self.update(iterable, **kwds)

    def __missing__(self, key):
        return 0

    def most_common(self, n=None):
        '''List the n most common elements and their counts from the most
        common to the least.  If n is None, then list all element counts.

        >>> Counter('abracadabra').most_common(3)
        [('a', 5), ('r', 2), ('b', 2)]

        '''
        if n is None:
            return sorted(self.iteritems(), key=itemgetter(1), reverse=True)
        return nlargest(n, self.iteritems(), key=itemgetter(1))

    def elements(self):
        '''Iterator over elements repeating each as many times as its count.

        >>> c = Counter('ABCABC')
        >>> sorted(c.elements())
        ['A', 'A', 'B', 'B', 'C', 'C']

        If an element's count has been set to zero or is a negative number,
        elements() will ignore it.

        '''
        for elem, count in self.iteritems():
            for _ in repeat(None, count):
                yield elem

    # Override dict methods where the meaning changes for Counter objects.

    @classmethod
    def fromkeys(cls, iterable, v=None):
        raise NotImplementedError(
            'Counter.fromkeys() is undefined.  Use Counter(iterable) instead.')

    def update(self, iterable=None, **kwds):
        '''Like dict.update() but add counts instead of replacing them.

        Source can be an iterable, a dictionary, or another Counter instance.

        >>> c = Counter('which')
        >>> c.update('witch')           # add elements from another iterable
        >>> d = Counter('watch')
        >>> c.update(d)                 # add elements from another counter
        >>> c['h']                      # four 'h' in which, witch, and watch
        4

        '''
        if iterable is not None:
            if hasattr(iterable, 'iteritems'):
                if self:
                    self_get = self.get
                    for elem, count in iterable.iteritems():
                        self[elem] = self_get(elem, 0) + count
                else:
                    dict.update(self, iterable)  # fast path when counter is empty
            else:
                self_get = self.get
                for elem in iterable:
                    self[elem] = self_get(elem, 0) + 1
        if kwds:
            self.update(kwds)

    def copy(self):
        'Like dict.copy() but returns a Counter instance instead of a dict.'
        return Counter(self)

    def __delitem__(self, elem):
        'Like dict.__delitem__() but does not raise KeyError for missing values.'
        if elem in self:
            dict.__delitem__(self, elem)

    def __repr__(self):
        if not self:
            return '%s()' % self.__class__.__name__
        items = ', '.join(map('%r: %r'.__mod__, self.most_common()))
        return '%s({%s})' % (self.__class__.__name__, items)

    # Multiset-style mathematical operations discussed in:
    #       Knuth TAOCP Volume II section 4.6.3 exercise 19
    #       and at http://en.wikipedia.org/wiki/Multiset
    #
    # Outputs guaranteed to only include positive counts.
    #
    # To strip negative and zero counts, add-in an empty counter:
    #       c += Counter()

    def __add__(self, other):
        '''Add counts from two counters.

        >>> Counter('abbb') + Counter('bcc')
        Counter({'b': 4, 'c': 2, 'a': 1})


        '''
        if not isinstance(other, Counter):
            return NotImplemented
        result = Counter()
        for elem in set(self) | set(other):
            newcount = self[elem] + other[elem]
            if newcount > 0:
                result[elem] = newcount
        return result

    def __sub__(self, other):
        ''' Subtract count, but keep only results with positive counts.

        >>> Counter('abbbc') - Counter('bccd')
        Counter({'b': 2, 'a': 1})

        '''
        if not isinstance(other, Counter):
            return NotImplemented
        result = Counter()
        for elem in set(self) | set(other):
            newcount = self[elem] - other[elem]
            if newcount > 0:
                result[elem] = newcount
        return result

    def __or__(self, other):
        '''Union is the maximum of value in either of the input counters.

        >>> Counter('abbb') | Counter('bcc')
        Counter({'b': 3, 'c': 2, 'a': 1})

        '''
        if not isinstance(other, Counter):
            return NotImplemented
        _max = max
        result = Counter()
        for elem in set(self) | set(other):
            newcount = _max(self[elem], other[elem])
            if newcount > 0:
                result[elem] = newcount
        return result

    def __and__(self, other):
        ''' Intersection is the minimum of corresponding counts.

        >>> Counter('abbb') & Counter('bcc')
        Counter({'b': 1})

        '''
        if not isinstance(other, Counter):
            return NotImplemented
        _min = min
        result = Counter()
        if len(self) < len(other):
            self, other = other, self
        for elem in ifilter(self.__contains__, other):
            newcount = _min(self[elem], other[elem])
            if newcount > 0:
                result[elem] = newcount
        return result


if __name__ == '__main__':
    import doctest

    print doctest.testmod()



try:
    from collections import OrderedDict
except ImportError:
    from qtop_py.legacy.ordereddict import OrderedDict

# example color maps
# order should be from more generic-->more specific
user_to_color_default = OrderedDict(
    [
        ('[a-z][_a-z0-9.-]*', 'Gray_L'),  # catch-all rule for many more names
        ('Atlassm', 'Red_L'),
        ('sgmatlas', 'Red_L'),
        ('patlas', 'Red_L'),
        ('satlas', 'Red_L'),
        ('satl', 'Red_L'),
        ('atlassgm', 'Red_L'),
        ('atlsgm', 'Red_L'),
        ('atlsg', 'Red_L'),
        ('atlasusr', 'Red_L'),
        ('atlass', 'Red_L'),
        ('laspt', 'Red_L'),
        ('atlpilot', 'Red_L'),
        ('atpilot', 'Red_L'),
        ('iatpilot', 'Red_L'),
        ('atlasplt', 'Red_L'),
        ('atlaspt', 'Red_L'),
        ('atlaspil', 'Red_L'),
        ('atlasplot', 'Red_L'),
        ('atlplot', 'Red_L'),
        ('atlpil', 'Red_L'),
        ('atlpl', 'Red_L'),
        ('atlaspilot', 'Red_L'),
        ('atlprod', 'Red'),
        ('atlasger', 'Red'),
        ('atlasde', 'Red'),
        ('datlas', 'Red'),
        ('atlasit', 'Red'),
        ('atlde', 'Red'),
        ('atlit', 'Red'),
        ('atlasprod', 'Red'),
        ('atlasprd', 'Red'),
        ('atlsprd', 'Red'),
        ('atlprd', 'Red'),
        ('atlpr', 'Red'),
        ('iatlas', 'Red'),
        ('iatprd', 'Red'),
        ('iatl', 'Red'),
        ('aatlpd', 'Red'),
        ('atlasfx', 'Red'),
        ('atlastw', 'Red'),
        ('atlx', 'Red'),
        ('atlp', 'Red'),
        ('atlu', 'Red'),
        ('atlhs', 'Red'),
        ('atlashs', 'Red'),
        ('atlasp', 'Red'),
        ('atlasfr', 'Red'),
        ('atlasana', 'Red'),
        ('Eatlas', 'Red'),
        ('shatlas', 'Red'),
        ('lcgatlas', 'Red'),
        ('atlasL', 'Red'),
        ('atlasil', 'Red'),
        ('atlanaly', 'Red'),
        ('atlas', 'Red'),
        ('atlbr', 'Red'),
        ('atls', 'Red'),
        ('atl', 'Red'),
        ('atfx', 'Red'),
        ('atprd', 'Red'),
        ('atcanpu', 'Red'),
        ('atcanpt', 'Red_L'),
        ('atcan', 'Red'),
        ('atcpu', 'Red'),
        ('atsgm', 'Red_L'),
        ('pilatlas', 'Red_L'),
        ('pilatl', 'Red_L'),
        ('zipilatl', 'Red_L'),
        ('patlit', 'Red_L'),
        ('platl', 'Red_L'),
        ('pltatlas', 'Red_L'),
        ('pltatl', 'Red_L'),
        ('atplt', 'Red_L'),
        ('atlplt', 'Red_L'),
        ('piatlas', 'Red_L'),
        ('piatla', 'Red_L'),
        ('piatl', 'Red_L'),
        ('sgmatl', 'Red_L'),
        ('zisgmatl', 'Red_L'),
        #( sgmatl, 'Red'),
        ('prdatlas', 'Red'),
        ('prdatl', 'Red'),
        ('prdat', 'Red'),
        ('ziprdatl', 'Red'),
        ('ziatlas', 'Red'),
        ('patls', 'Red'),
        ('patlit', 'Red'),
        ('patl', 'Red'),
        ('nordugrid-atlas', 'Red'),
        ('nlv', 'Red_L'),
        ('ncfk', 'Green_L'),
        ('usatlas', 'Red'),
        #( CMS VO commonly found pool account names)
        ('cmssgm', 'Green_L'),
        ('cmsplt', 'Green_L'),
        ('pltcms', 'Green_L'),
        ('cmspilot', 'Green_L'),
        ('cmspil', 'Green_L'),
        ('pilcms', 'Green_L'),
        ('pcms', 'Green_L'),
        ('sgmcms', 'Green_L'),
        ('sgmcm', 'Green_L'),
        ('scms', 'Green_L'),
        ('priocms', 'Green_L'),
        ('cmsprio', 'Green_L'),
        ('pricms', 'Green_L'),
        ('cmsprod', 'Green'),
        ('cmsprd', 'Green'),
        ('cmsmcp', 'Green'),
        ('cmsnu', 'Green'),
        ('cmst1prd', 'Green'),
        ('cmprd', 'Green'),
        ('uscmsPool', 'Green'),
        ('uscms', 'Green'),
        ('cmsp', 'Green'),
        ('cmsusr', 'Green'),
        ('cmsuwu', 'Green'),
        ('cmsana', 'Green'),
        ('cmsger', 'Green'),
        ('cmss', 'Green'),
        ('cmszu', 'Green'),
        ('cmsau', 'Green'),
        ('twcms', 'Green'),
        ('cmsmu', 'Green'),
        ('dcms', 'Green'),
        ('prdcms', 'Green'),
        ('icms', 'Green'),
        ('cms', 'Green'),
        #( ALICE VO commonly found pool account name)s
        ('ali', 'Cyan'),
        ('alis', 'Cyan'),
        ('alice', 'Cyan'),
        ('alisc', 'Cyan'),
        ('alisgm', 'Cyan'),
        ('alisg', 'Cyan'),
        ('alibs', 'Cyan'),
        ('alikn', 'Cyan'),
        ('sali', 'Cyan'),
        ('ialice', 'Cyan'),
        ('salice', 'Cyan'),
        ('caliceuser', 'Cyan'),
        ('caliceusr', 'Cyan'),
        ('calic', 'Cyan'),
        ('calice', 'Cyan'),
        ('sgmali', 'Cyan'),
        ('sgmalice', 'Cyan'),
        ('alicesgm', 'Cyan'),
        #( LHCb VO commonly found pool account names)
        ('pdlhcb', 'Pink'),
        ('prdlhcb', 'Pink'),
        ('prdlhb', 'Pink'),
        ('prdlhc', 'Pink'),
        ('lhcbprd', 'Pink'),
        ('lhbprd', 'Pink'),
        ('lhprd', 'Pink'),
        ('ilhcb', 'Pink'),
        ('lhcbsgm', 'Purple'),
        ('lhcbs', 'Purple'),
        ('sgmlhcb', 'Purple'),
        ('sgmlhb', 'Purple'),
        ('sgmlhc', 'Purple'),
        ('lhcbplt', 'Purple'),
        ('lhcbpilot', 'Purple'),
        ('lhpilot', 'Purple'),
        ('lhcpilot', 'Purple'),
        ('lhcbpil', 'Purple'),
        ('lhbpil', 'Purple'),
        ('pillhcb', 'Purple'),
        ('plhcb', 'Purple'),
        ('pilhcb', 'Purple'),
        ('pltlhcb', 'Purple'),
        ('pillhb', 'Purple'),
        ('pllhc', 'Purple'),
        ('tlhcb', 'Pink'),
        ('lhcbhs', 'Pink'),
        ('lhcbp', 'Pink'),
        ('lhcb', 'Pink'),
        ('lhcp', 'Pink'),
        #( dteam VO commonly found pool account name)s
        ('dteamsgm', 'Brown'),
        ('dteamprd', 'Brown'),
        ('dteamuser', 'Brown'),
        ('dteamusr', 'Brown'),
        ('dteam', 'Brown'),
        ('dte', 'Brown'),
        #( OPS VO commonly found pool account names)
        ('opsplt', 'Yellow'),
        ('opsusr', 'Yellow'),
        ('opssgm', 'Yellow'),
        ('opssg', 'Yellow'),
        ('opsgm', 'Yellow'),
        ('sgmops', 'Yellow'),
        ('zisgmops', 'Yellow'),
        ('opsprd', 'Yellow'),
        ('opspil', 'Yellow'),
        ('pilops', 'Yellow'),
        ('samgrid', 'Yellow'),
        ('opss', 'Yellow'),
        ('sops', 'Yellow'),
        ('opsiber', 'Yellow'),
        ('opsib', 'Yellow'),
        ('ops', 'Yellow'),
        #( Other VOs from the EGEE-I,II,III era)
        ('egee', 'Blue_L'),
        #( Biomed VO)
        ('bio', 'Blue_L'),
        ('biocw', 'Blue_L'),
        ('biome', 'Blue_L'),
        ('biomd', 'Blue_L'),
        ('biomed', 'Blue_L'),
        ('biomedusr', 'Blue_L'),
        ('biostats', 'Blue_L'),
        ('biotech', 'Blue_L'),
        #( Gear V)O
        ('gearsgm', 'Blue_L'),
        ('gearprd', 'Blue_L'),
        ('gear', 'Blue_L'),
        #( DECH V)O
        ('dechsgm', 'Blue_L'),
        ('dechprd', 'Blue_L'),
        ('dechusr', 'Blue_L'),
        ('dech', 'Blue_L'),
        #( SEE V)O
        ('seeops', 'Blue_L'),
        ('seops', 'Blue_L'),
        ('seegrid', 'Blue_L'),
        ('seevo', 'Blue_L'),
        ('see', 'Blue_L'),
        #( ESR V)O
        ('esrsgm', 'Blue_L'),
        ('esr', 'Blue_L'),
        ('earthscience', 'Blue_L'),
        #( Fusion, auvergrid, compchem, enmr, voce, gaussian, balticgrid),
        #( Digital Media VO)s
        ('fusionprd', 'Blue_L'),
        ('fusionsgm', 'Blue_L'),
        ('fusion', 'Blue_L'),
        ('fusio', 'Blue_L'),
        ('fusi', 'Blue_L'),
        ('fusn', 'Blue_L'),
        ('fus', 'Blue_L'),
        ('auvergrid', 'Blue_L'),
        ('enmr', 'Blue_L'),
        ('complex', 'Blue_L'),
        ('compchem', 'Blue_L'),
        ('compc', 'Blue_L'),
        ('compl', 'Blue_L'),
        ('cmplx', 'Blue_L'),
        ('vocesgm', 'Blue_L'),
        ('voceprd', 'Blue_L'),
        ('voce', 'Blue_L'),
        ('gaussian', 'Blue_L'),
        ('balticgrid', 'Blue_L'),
        ('digmedia', 'Blue_L'),
        ('dmedia', 'Blue_L'),
        #( Hone VO)
        ('prdhone', 'Cyan_L'),
        ('prdhne', 'Cyan_L'),
        ('prdhon', 'Cyan_L'),
        ('honecker', 'Cyan_L'),
        ('honeprd', 'Cyan_L'),
        ('honesgm', 'Cyan_L'),
        ('phone', 'Cyan_L'),
        ('hone', 'Cyan_L'),
        ('honp', 'Cyan_L'),
        #( Other (High Energy) Physics VOs)
        ('sixto', 'Cyan_L'),
        ('sixt', 'Cyan_L'),
        ('babaradm', 'Cyan_L'),
        ('babarpro', 'Cyan_L'),
        ('babar', 'Cyan_L'),
        ('pheno', 'Cyan_L'),
        ('dzerojim', 'Cyan_L'),
        ('dzero', 'Cyan_L'),
        ('dze', 'Cyan_L'),
        ('dzerqa', 'Cyan_L'),
        ('theophys_', 'Cyan_L'),
        ('theophys', 'Cyan_L'),
        ('zeususr', 'Cyan_L'),
        ('zeus', 'Cyan_L'),
        ('argo', 'Cyan_L'),
        ('ilcprd', 'Cyan_L'),
        ('ilcpr', 'Cyan_L'),
        ('ilcp', 'Cyan_L'),
        ('ilcusr', 'Cyan_L'),
        ('ilcger', 'Cyan_L'),
        ('ilc', 'Cyan_L'),
        ('prdilc', 'Cyan_L'),
        ('pilc', 'Cyan_L'),
        ('sgmilc', 'Cyan_L'),
        ('augersgm', 'Cyan_L'),
        ('augerprd', 'Cyan_L'),
        ('auger', 'Cyan_L'),
        ('augp', 'Cyan_L'),
        ('aug', 'Cyan_L'),
        ('chatlas', 'Cyan_L'),
        ('chcms', 'Cyan_L'),
        ('chlhcb', 'Cyan_L'),
        ('geant', 'Cyan_L'),
        ('lhcft', 'Cyan_L'),
        ('magic', 'Cyan_L'),
        ('scier', 'Cyan_L'),
        ('planck', 'Cyan_L'),
        ('kaust', 'Brown'),
        #( Extras; these are really randomly found user names; needed for screen)
        #( clarity while in color more)
        ('j', 'Yellow'),       ## yes, let's start editing the colormap file around here ;-)
        ('mwilli', 'Brown'),
        ('por', 'Blue'),
        ('dorisf', 'Cyan'),
        ('campoman', 'Brown'),
        ('gallet', 'Blue'),
        ('tlemmin', 'Cyan'),
        ('ls', 'Cyan'),
        ('train', 'Blue_L'),
        ('pkoro', 'Brown'),
        ('fotis', 'Blue'),
        ('fgeorgatos', 'Blue'),
        ('fgeor', 'Blue'),
        ('svc', 'Cyan'),
        ('svc_hpc', 'Green'),
        ('dashed', 'Gray_L'),
        ('purple', 'Purple'),
        ('Red', 'Red'),
        ('DGray', 'Gray_D'),
        ('Brown', 'Brown'),
        ('_', 'White'),
        ('#', 'Gray_D'),
        ('account_not_colored', 'reset'),
        ('NoPattern', 'White'),
    ]
)

color_to_code = {
    'Gray_D': '1;30',
    'Gray_L': '1;37',
    'White': '1;37',
    'Red': '0;31',
    'Red_L': '1;31',
    'Red_LOnGrayBG': '1;31;40',
    'Green': '0;32',
    'Green_L': '1;32',
    'Purple': '0;35',
    'PurpleOnGrayBG': '0;35;40',
    'Pink': '1;35',
    'Brown': '0;33',
    'Yellow': '1;33',
    'Cyan': '0;36',
    'Cyan_L': '1;36',
    'Blue': '0;34',
    'Blue_L': '1;34',
    'reset': '0',
    'GrayBG': ';40',
    'MaroonBG': ';41',
    'GreenBG': ';42',
    'GoldBG': ';43',
    'BlueBG': ';44',
    'MagentaBG': ';45',
    'BlackOnBlack': '0;30;40',
    'RedOnBlack': '0;31;40',
    'GreenOnBlack': '0;32;40',
    'OrangeOnBlack': '0;33;40',
    'BlueOnBlack': '0;34;40',
    'PinkOnBlack': '0;35;40',
    'CyanOnBlack': '0;36;40',
    'GrayOnBlack': '0;37;40',
    'BlackOnRed': '0;30;41',
    'RedOnRed': '0;31;41',
    'GreenOnRed': '0;32;41',
    'OrangeOnRed': '0;33;41',
    'BlueOnRed': '0;34;41',
    'PinkOnRed': '0;35;41',
    'CyanOnRed': '0;36;41',
    'GrayOnRed': '0;37;41',
    'BlackOnGreen': '0;30;42',
    'RedOnGreen': '0;31;42',
    'GreenOnGreen': '0;32;42',
    'OrangeOnGreen': '0;33;42',
    'BlueOnGreen': '0;34;42',
    'PinkOnGreen': '0;35;42',
    'CyanOnGreen': '0;36;42',
    'GrayOnGreen': '0;37;42',
    'BlackOnOrange': '0;30;43',
    'RedOnOrange': '0;31;43',
    'GreenOnOrange': '0;32;43',
    'OrangeOnOrange': '0;33;43',
    'BlueOnOrange': '0;34;43',
    'PinkOnOrange': '0;35;43',
    'CyanOnOrange': '0;36;43',
    'GrayOnOrange': '0;37;43',
    'BlackOnBlue': '0;30;44',
    'RedOnBlue': '0;31;44',
    'GreenOnBlue': '0;32;44',
    'OrangeOnBlue': '0;33;44',
    'BlueOnBlue': '0;34;44',
    'PinkOnBlue': '0;35;44',
    'CyanOnBlue': '0;36;44',
    'GrayOnBlue': '0;37;44',
    'BlackOnPink': '0;30;45',
    'RedOnPink': '0;31;45',
    'GreenOnPink': '0;32;45',
    'OrangeOnPink': '0;33;45',
    'BlueOnPink': '0;34;45',
    'PinkOnPink': '0;35;45',
    'CyanOnPink': '0;36;45',
    'GrayOnPink': '0;37;45',
    'BlackOnCyan': '0;30;46',
    'RedOnCyan': '0;31;46',
    'GreenOnCyan': '0;32;46',
    'OrangeOnCyan': '0;33;46',
    'BlueOnCyan': '0;34;46',
    'PinkOnCyan': '0;35;46',
    'CyanOnCyan': '0;36;46',
    'GrayOnCyan': '0;37;46',
    'BlackOnGray': '0;30;47',
    'RedOnGray': '0;31;47',
    'GreenOnGray': '0;32;47',
    'OrangeOnGray': '0;33;47',
    'BlueOnGray': '0;34;47',
    'PinkOnGray': '0;35;47',
    'CyanOnGray': '0;36;47',
    'GrayOnGray': '0;37;47',
    'CyanBG': ';46',
    'Gray2BG': ';47',
    '': '',
    'NOBG': '',
}

queue_to_color = OrderedDict(
    [
        ('Pending', 'Yellow'),
        ('whole', 'Purple'),
        (r'grid\d+M', 'Red_L'),
        ('dteam.*', 'Red_L'),
        ('alice', 'Red_L'),
        ('cta', 'Pink'),
        ('grid3000M', 'Purple'),
        ('grid500M', 'Pink'),
        ('grid1000M', 'Red_L'),
        ('grid2000M', 'Red'),
        ('grid3000M', 'Blue'),
        ('grid4000M', 'Yellow'),
        ('transfer', 'Yellow'),
        ('urgent', 'Red'),
    ]
)

nodestate_to_color_default = OrderedDict(
    [
        ('hqw', 'PurpleOnGrayBG'),
        ('r', 'Blue'),
        ('d', 'Gray'),
        ('o', 'MaroonBG'),
    ]
)



class BaseViewport(object):
    """
    Class that implements a basic viewport [window into the qtop matrix] and basic movement restrictions
    """
    def __init__(self, vstart=0, hstart=0):
        self._v_start = vstart
        self._v_stop = 0
        self._h_start = hstart
        self._h_stop = 0
        self.max_height = 0  # zero until a non-empty qtop output file is read
        self.max_width = 0
        self.v_term_size = 0
        self.h_term_size = 0

    def get_up_limit(self):
        return 0

    def get_right_limit(self):
        # TODO check validity of this: extreme case for no initial max_width, i.e. no matrix output
        return max(self.max_width - self.h_term_size, 0)

    def get_down_limit(self):
        # TODO check validity of this: extreme case for no initial max_height, i.e. no matrix output
        return max(self.max_height - self.v_term_size, 0)

    def get_left_limit(self):
        return 0

    def would_cross_up_limit(self, value):
        return value < self.get_up_limit()

    def would_cross_right_limit(self, value):
        return value > self.get_right_limit()

    def would_cross_down_limit(self, value):
        return value > self.get_down_limit()

    def would_cross_left_limit(self, value):
        return value < self.get_left_limit()


class Viewport(BaseViewport):
    """
    Class that extends BaseViewport to provide the actual movements of the Viewport
    """
    @property
    def h_start(self):
        assert not (self.would_cross_left_limit(self._h_start) and self.would_cross_right_limit(self._h_start))
        return self._h_start

    @h_start.setter
    def h_start(self, value):
        if self.would_cross_right_limit(value):
            self._h_start = self.get_right_limit()
        elif self.would_cross_left_limit(value):
            self._h_start = self.get_left_limit()
        else:
            self._h_start = value

    @property
    def v_start(self):
        assert not (self.would_cross_up_limit(self._v_start) and self.would_cross_down_limit(self._v_start))
        return self._v_start

    @v_start.setter
    def v_start(self, value):
        if self.would_cross_up_limit(value):
            self._v_start = self.get_up_limit()
        elif self.would_cross_down_limit(value):
            self._v_start = self.get_down_limit()
        else:
            self._v_start = value

    @property
    def h_stop(self):
        return self._h_start + self.h_term_size

    @property
    def v_stop(self):
        return self._v_start + self.v_term_size

    def set_term_size(self, term_height, term_columns):
        self.v_term_size = term_height
        self.h_term_size = term_columns

    def get_term_size(self):
        return self.v_term_size, self.h_term_size

    def scroll_down(self):
        success = False
        if self.v_stop < self.max_height:
            self.v_start += self.v_term_size
            success = True
        return success

    def scroll_bottom(self):
        success = False
        if self.v_stop < self.max_height:
            self.v_start = self.get_down_limit()
            success = True
        return success

    def scroll_up(self):
        self.v_start -= self.v_term_size
        return True

    def scroll_top(self):
        self.v_start = 0
        return True

    def scroll_right(self):
        if self.h_start + self.h_term_size >= self.max_width:
            return False
        self.h_start += self.h_term_size / 2
        return True

    def scroll_far_right(self):
        self.h_start = self.get_right_limit()

    def scroll_left(self):
        self.h_start -= self.h_term_size / 2

    def scroll_far_left(self):
        self.h_start = 0

    def reset_display(self):
        self.v_start = 0
        self.h_start = 0



from qtop_py.plugins import pbs
import pytest


@pytest.mark.parametrize('core_selections, result',
     (
             ('0-4,30-31', ["0","1","2","3","4","30","31"]),
             ('5-9', ["5","6","7","8","9"]),
             ('0,3,5-7,11', ["0","3","5","6","7","11"]),
             ('1,2-4,5-6,9', ["1","2","3","4","5","6","9"])
     ),
)
def test_get_corejob_from_range(core_selections, result, job=None):
    result = iter(result)
    for (core, job) in pbs.PBSBatchSystem.get_corejob_from_range(core_selections, job):
        assert core == next(result)



import re
import sys
from itertools import count


class StatExtractor(object):
    """
    Extracts qstat/qstatq information from files coming from the respective batch system commmands
    (PBS, OAR, SGE etc)
    """

    def __init__(self, config, options):
        self.config = config
        self.options = options
        self.anonymize = self.anonymize_func()

    def _process_qstat_line(self, re_search, line, re_match_positions):
        """
        extracts data from a tabular qstat-like file
        returns a list
        """
        qstat_values = dict()
        m = re.search(re_search, line.strip())

        try:
            job_id, user, job_state, queue = [m.group(x) for x in re_match_positions]
        except AttributeError:
            sys.__stdout__.write('Line: %s not properly parsed by regex expression.\n' % line.strip())
            raise
        job_id = job_id.split('.')[0]
        user = user if not self.options.ANONYMIZE else self.anonymize(user, 'users')
        for key, value in [('JobId', job_id), ('UnixAccount', user), ('S', job_state), ('Queue', queue)]:
            qstat_values[key] = value
        return qstat_values

    def anonymize_func(self):
        """
        creates and returns an _anonymize_func object (closure)
        Anonymisation can be used by the user for providing feedback to the developers, without leaking cluster's private data.
        i.e. the logs and the output should no longer contain sensitive information about the clusters ran by the user.
        """
        counters = {}
        stored_dict = {}
        for key in ['users', 'wns', 'qs']:
            counters[key] = count()

        maps = {
            'users': '_anon_user_',
            'wns': '_anon_wn_',
            'qs': '_anon_q_'
        }

        def _anonymize_func(s, a_type):
            """
            d4-p4-04 --> d_anon_wn_0
            d4-p4-05 --> d_anon_wn_1
            biomed017--> b_anon_user_0
            alice    --> a_anon_q_0
            """
            dup_counter = counters[a_type]

            s_type = maps[a_type]
            cnt = '0'
            new_name_parts = [s[0], s_type, cnt]
            if s not in stored_dict:
                cnt = str(dup_counter.next())
                new_name_parts.pop()
                new_name_parts.append(cnt)
            stored_dict.setdefault(s, (''.join(new_name_parts), s_type))
            return stored_dict[s][0]

        return _anonymize_func


class GenericBatchSystem(object):
    def __init__(self):
        pass

    def get_queues_info(self):
        raise NotImplementedError

    def get_worker_nodes(self, job_ids, job_queues, options):
        raise NotImplementedError

    def get_jobs_info(self, qstats):
        raise NotImplementedError

    @staticmethod
    def get_mnemonic():
        raise NotImplementedError

    @staticmethod
    def ensure_worker_nodes_have_qnames(_worker_nodes, job_ids, job_queues):
        """
        This gets the queues associated with each worker node.
        SGE systems already contain this information.
        """
        if not _worker_nodes:
            return _worker_nodes
        job_ids_queues = dict(zip(job_ids, job_queues))
        for worker_node in _worker_nodes:
            my_jobs = worker_node['core_job_map'].values()
            my_queues = set(job_ids_queues[re.sub(r'\[\d+\]', r'[]', job_id)] for job_id in my_jobs)  # also for job arrays
            worker_node['qname'] = list(my_queues)
        return _worker_nodes



from setuptools import setup
from qtop_py import __version__

import sys

if sys.version_info < (2, 5):
    sys.exit('Sorry, Python < 2.5 is not supported')

f = open("README.rst", 'r')
try:
    long_description = f.read()
finally:
    f.close()

setup(
    name='qtop',
    version=__version__,
    description="""qtop: the fast text mode way to monitor your cluster's utilization and status;
      the time has come to take back control of your cluster's scheduling business""",
    license="MIT",
    long_description=long_description,
    author='Sotiris Fragkiskos',
    author_email='sfranky@gmail.com',
    url="https://github.com/qtop/qtop",
    packages=['qtop_py',
              'qtop_py.legacy',
              'qtop_py.plugins',
              'qtop_py.ui'
              ],
    package_dir={'qtop_py': 'qtop_py'},
    package_data={'qtop_py': ['../qtopconf.yaml', '../qtop.py']},
    scripts=['qtop']
)



import logging
import os
import errno
import tempfile
import tarfile
import sys
import glob
import datetime


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def check_empty_file(orig_file):
    if os.path.getsize(orig_file) == 0:
        raise FileEmptyError(orig_file)


def get_new_temp_file(_savepath, suffix, prefix):  # **kwargs
    """
    Using mkstemp instead of NamedTemporaryFile because a file descriptor
    is needed to redirect sys.stdout to.
    """
    fd, temp_filepath = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=_savepath)  # **kwargs
    logging.debug('temp_filepath: %s' % temp_filepath)
    # out_file = os.fdopen(fd, 'w')
    return fd, temp_filepath


def safe_exit_with_file_close(handle, name, stdout, options, _savepath,
                              qtop_logfile, sample_filename, delete_file=False):
    sys.stdout.write('\nExiting. Thank you for ..watching ;)\n')
    sys.stdout.flush()
    sys.stdout.close()
    try:
        os.close(handle)
    except OSError:
        pass
    if delete_file:
        os.unlink(name)  # this deletes the file
    # sys.stdout = stdout
    if options.SAMPLE >= 1:
        add_to_sample([qtop_logfile], _savepath, sample_filename)
    sys.exit(0)


def init_sample_file(options, _savepath, SAMPLE_FILENAME, scheduler_output_filenames, QTOPCONF_YAML, QTOPPATH):
    """
    If the user wants to give feedback to the developers for a bugfix via the -L cmdline switch,
    this initialises a tar file, and adds:
    * the scheduler output files (-L),
    * and source files (-LL)
    to the tar file
    """
    if options.SAMPLE >= 1:
        # clears any preexisting tar files
        tar_out = tarfile.open(os.path.join(_savepath, SAMPLE_FILENAME), mode='w')
        tar_out.close()
        # add all scheduler output files to sample
        [add_to_sample([scheduler_output_filenames[fn]], _savepath, SAMPLE_FILENAME)
         for fn in scheduler_output_filenames if os.path.isfile(scheduler_output_filenames[fn])]

    if options.SAMPLE >= 2:
        add_to_sample([os.path.join(os.path.realpath(QTOPPATH), QTOPCONF_YAML)], _savepath, SAMPLE_FILENAME)
        source_files = glob.glob(os.path.join(os.path.realpath(QTOPPATH), '*.py'))
        add_to_sample(source_files, _savepath, SAMPLE_FILENAME, subdir='qtop_py')


def add_to_sample(filepaths_to_add, _savepath, sample_file, sample_method=tarfile, subdir=None):
    """
    opens sample_file in path savepath and adds files filepaths_to_add
    """
    assert isinstance(filepaths_to_add, list)
    sample_out = sample_method.open(os.path.join(_savepath, sample_file), mode='a')
    for filepath_to_add in filepaths_to_add:
        path, fn = filepath_to_add.rsplit('/', 1)
        try:
            logging.debug('Adding %s to sample...' % filepath_to_add)
            sample_out.add(filepath_to_add, arcname=fn if not subdir else os.path.join(subdir, fn))
        except tarfile.TarError:  # TODO: test what could go wrong here
            logging.error('There seems to be something wrong with the tarfile. Skipping...')
    else:
        logging.debug('Closing sample...')
        sample_out.close()


def get_sample_filename(SAMPLE_FILENAME, config):
    if config['overwrite_sample_file']:
        SAMPLE_FILENAME = SAMPLE_FILENAME % {'datetime': ''}
    else:
        SAMPLE_FILENAME = SAMPLE_FILENAME % {'datetime': '_' + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")}
    return SAMPLE_FILENAME


class FileNotFound(Exception):
    def __init__(self, fn):
        msg = "File %s not found.\nMaybe the correct scheduler is not specified?" % fn
        Exception.__init__(self, msg)
        logging.critical(msg)
        self.fn = fn


class FileEmptyError(Exception):
    def __init__(self, fn):
        msg = "File %s is empty.\n" \
              "Is your batch scheduler loaded with jobs?" % fn
        Exception.__init__(self, msg)
        logging.warning(msg)
        self.fn = fn


def deprecate_old_output_files(config):
    """
    deletes older json and .out files in savepath directory.
    """
    time_alive = get_timedelta(parse_time_input(config['auto_delete_old_output_files_after']))
    _savepath = config['savepath']
    for f in os.listdir(_savepath):
        if (not f.endswith(('json', '.out'))) or f.endswith('rec.out'):
            continue
        curpath = os.path.join(_savepath, f)
        file_modified = datetime.datetime.fromtimestamp(os.path.getmtime(curpath))
        if datetime.datetime.now() - file_modified > time_alive:
            os.remove(curpath)


def get_timedelta(extra_kw_args):
    """
    This solely exists to allow timedelta.timedelta's keyword argument (minutes/seconds/hours=...)
    to be selected with a variable.
    """
    return datetime.timedelta(**extra_kw_args)


def parse_time_input(_time):
    """
    the func accepts a _time str in either (h)ours, (m)inutes, or (s)econds, using the respective suffix,
    e.g. '5h', or '10m', or '30s'
    A tuple is returned, e.g. (5, 'hours')
    """
    assert _time.endswith(('h', 'm', 's'))
    try:
        int(_time[:-1])
    except ValueError:
        logging.critical('Time input given must be a number followed by the letter h/m/s. Exiting.')

    quantity, user_unit_suffix = _time[:-1], _time[-1]
    units = {'m': 'minutes', 's': 'seconds', 'h': 'hours'}
    user_unit = units[user_unit_suffix]

    return {user_unit: int(quantity)}



import logging
import sys
from optparse import OptionParser
import fileutils
from qtop_py.colormap import *
from qtop_py.constants import QTOP_LOGFILE


def init_logging(options):
    if not options.verbose:
        log_level = logging.WARN
    elif options.verbose == 1:
        log_level = logging.INFO
    elif options.verbose >= 2:
        log_level = logging.DEBUG

    fileutils.mkdir_p(QTOP_LOGFILE.rsplit('/', 1)[0])  # logfile path

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    if options.verbose >= 3:
        # this prepends date/time
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s', "%Y-%m-%d %H:%M:%S")
    else:
        formatter = logging.Formatter('%(levelname)s - %(message)s')

    fh = logging.FileHandler(QTOP_LOGFILE)
    fh.setLevel(log_level)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    fh = logging.StreamHandler()
    fh.setLevel(logging.ERROR)
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    logger.disabled = False  # TODO: maybe make this a cmdline switch? -D ?

    logging.info("\n" + "=" * 50 + "STARTING NEW LOG ENTRY..." + "=" * 50 + "\n\n")

    logging.debug('Verbosity level = %s' % options.verbose)
    logging.debug("input, output isatty: %s\t%s" % (sys.stdin.isatty(), sys.stdout.isatty()))


def parse_qtop_cmdline_args():
    parser = OptionParser()  # for more details see http://docs.python.org/library/optparse.html

    parser.add_option("-1", "--disablesection1", action="store_true", dest="sect_1_off", default=False,
                      help="Disable first section of qtop, i.e. Job Accounting Summary")
    parser.add_option("-2", "--disablesection2", action="store_true", dest="sect_2_off", default=False,
                      help="Disable second section of qtop, i.e. Worker Node Occupancy")
    parser.add_option("-3", "--disablesection3", action="store_true", dest="sect_3_off", default=False,
                      help="Disable third section of qtop, i.e. User Accounts and Pool Mappings")
    parser.add_option("-a", "--blindremapping", action="store_true", dest="BLINDREMAP", default=False,
                      help="This may be used in situations where node names are not a pure arithmetic seq "
                           "(e.g. rocks clusters)")
    # TODO . Must also anonymise input files, or at least exclude them from the tarball.
    parser.add_option("-A", "--anonymize", action="store_true", dest="ANONYMIZE", default=False,
                      help="Masks unix account names and workernode names for security reasons (sending bug reports etc)."
                           "Temporarily NOT to be used, as scheduler input files are not anonymised yet.")
    parser.add_option("-b", "--batchSystem", action="store", dest="BATCH_SYSTEM", default=None)
    parser.add_option("-c", "--COLOR", action="store", dest="COLOR", default="AUTO", choices=['ON', 'OFF', 'AUTO'],
                      help="Enable/Disable color in qtop output. AUTO detects tty (for watch -d)")
    parser.add_option("-C", "--classic", action="store_true", dest="CLASSIC", default=False,
                      help="tries to mimic legacy qtop display as much as possible")
    parser.add_option("-d", "--debug", action="store_true", dest="DEBUG", default=False,
                      help="print debugging messages in stdout, not just in the log file.")
    parser.add_option("-E", "--export", action="store_true", dest="EXPORT", default=False,
                      help="export cluster data to json")
    parser.add_option("-F", "--ForceNames", action="store_true", dest="FORCE_NAMES", default=False,
                      help="force names to show up instead of numbered WNs even for very small numbers of WNs")
    parser.add_option("-f", "--setCUSTOMCONFFILE", action="store", type="string", dest="CONFFILE")
    parser.add_option("-G", "--get_GECOS_via_getent_passwd", action="store_true", dest="GET_GECOS", default=False,
                      help="get user details by issuing getent passwd for all users mentioned in qtop input files.")
    parser.add_option("-m", "--noMasking", action="store_true", dest="NOMASKING", default=False,
                      help="Don't mask early empty WNs (default: if the first 30 WNs are unused, counting starts from 31).")
    parser.add_option("-o", "--option", action="append", dest="OPTION", type="string", default=[],
                      help="Override respective option in QTOPCONF_YAML file")
    parser.add_option("-O", "--onlysavetofile", action="store_true", dest="ONLYSAVETOFILE", default=False,
                      help="Do not print results to stdout")
    parser.add_option("-r", "--removeemptycorelines", dest="REM_EMPTY_CORELINES", action="count", default=False,
                      help="If a whole row consists of not-really-there ('#') core lines, remove the row."
                           "If doubled (-rr), remove the row even if it also consists of free, unused cores ('_').")
    parser.add_option("-R", "--replay", action="callback", dest="REPLAY", callback=_watch_callback,
                      help="instant replay from a specific moment in time for the "
                           "cluster, and for a specified duration. The value "
                           "provided should be in either of the following formats: "
                           "yyyymmddTHHMMSS, e.g. 20161118T182300, (explicit form) "
                           "HHMM, e.g. 1823 (current day is implied),\t\t "
                           "mmddTHHMM, e.g. 1118 T1823(current year is implied).  "
                           "A second value is optional and denotes the desired "
                           "length of the playback, e.g. -R 1823 1m, "
                           "or -R 1800 1h. A default duration of 2m is used, if"
                           "no value is given.")
    parser.add_option("-s", "--SetSourceDir", dest="SOURCEDIR",
                      help="Set the source directory where the batch scheduler output files reside")
    parser.add_option("-S", "--StrictCheck", dest="STRICTCHECK", action="store_true",
                      help="Do a check on the quality of the scheduler output by comparing "
                           "the reported total running jobs against the actual ones found/displayed in qtop")
    parser.add_option("-T", "--Transpose", dest="TRANSPOSE", action="store_true", default=False,
                      help="Rotate matrices' positioning by 90 degrees")
    parser.add_option("-B", "--web", dest="WEB", action="store_true", default=False,
                      help="Enable web interface in 8080")
    parser.add_option("-v", "--verbose", dest="verbose", action="count",
                      help="Increase verbosity (specify multiple times for more)")
    parser.add_option("-V", "--version", dest="version", action="store_true",
                      help="Print qtop version")
    parser.add_option("-w", "--watch", dest="WATCH", action="callback", callback=_watch_callback,
                      help="Mimic shell's watch behaviour. Use with optional argument, e.g. '-w 10' to refresh every 10 sec"
                           "instead of the default which is 2 sec.")
    parser.add_option("-L", "--sample", action="count", dest="SAMPLE", default=False,
                      help="Create a sample file. A single L creates a tarball with the log, scheduler output files, "
                           "qtop output. Two L's additionaly include the qtop_conf yaml file, and qtop qtop_py.")
    # parser.add_option("-f", "--setCOLORMAPFILE", action="store", type="string", dest="COLORFILE")  # TODO

    (options, args) = parser.parse_args()
    return options, args


def _watch_callback(option, opt_str, value, parser):
    """
    This is the official example from optparse for variable arguments
    """
    assert value is None
    value = []

    def is_floatable(str):
        try:
            float(str)
            return True
        except ValueError:
            return False

    for arg in parser.rargs:
        # stop on --foo like options
        if arg[:2] == "--" and len(arg) > 2:
            break
        # stop on -a, but not on -3 or -3.0
        if arg[:1] == "-" and len(arg) > 1 and not is_floatable(arg):
            break
        value.append(arg)
    if not value:  # zero arguments!
        value.append(0)
    else:
        del parser.rargs[:len(value)]
    setattr(parser.values, option.dest, value)


class ColorStr(object):
    """
    ColorStr instances are normal strings with color information attached to them,
    to be used with colorize(), e.g.
    print colorize(s.str, color_func=s.color)
    print colorize(s, mapping=nodestate_to_color, pattern=s.initial)
    """
    def __init__(self, string='', color=''):
        self.str = string
        self.color = color
        self.initial = self.str[0]
        self.index = 0
        self.stop = len(self.str)

    def __str__(self):
        return str(self.str)

    def __repr__(self):
        return repr(self.str)

    def __len__(self):
        return len(self.initial)

    def __iter__(self):
        return self

    def next(self):
        if self.index == self.stop:
            raise StopIteration
        self.index += 1
        # return self.initial
        return self

    def __contains__(self, item):
        return item in self.str

    def __equals__(self, other):
        return self.str == other.str

    @classmethod
    def from_other_color_str(cls, color_str):
        return cls(string=color_str.str)


class CountCalls(object):
    """
    Decorator that keeps track of the number of times a function is called.
    """

    __instances = {}

    def __init__(self, f):
        self.__f = f
        self.__numcalls = 0
        CountCalls.__instances[f] = self

    def __call__(self, *args, **kwargs):
        self.__numcalls += 1
        return self.__f(*args, **kwargs)

    def count(self):
        "Return the number of times the function f was called."
        return CountCalls.__instances[self.__f].__numcalls

    @staticmethod
    def counts():
        "Return a dict of {function: # of calls} for all registered functions."
        return dict([(f.__name__, CountCalls.__instances[f].__numcalls) for f in CountCalls.__instances])
